/*
 *	wcollector, wireless event collector for mac80211 linux kernel module
 *	Copyright (C) 2011  Carlo Vallati (c.vallati@iet.unipi.it)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version 2
 *	of the License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 *	02110-1301, USA.
 */

#include <netlink/netlink.h>
#include <netlink/genl/genl.h>
#include <netlink/genl/ctrl.h>
#include <netlink/genl/family.h>
#include <stdint.h>
#include <getopt.h>

#include "wcollector.h"
#include "ieee80211.h"

#define VERSION_STR "\"0.1\""

struct nl_sock *sock;
struct nl_msg *msg;
struct nl_cb *cb;
struct nl_cache *cache;
struct genl_family *family;

struct mac_address {
        unsigned char addr[6];
};

/*
 * 	Callback function for event collected from the kernel
 */

static int process_messages_cb(struct nl_msg *msg, void *arg)
{
	struct nlattr *attrs[COLLECTOR_ATTR_MAX+1];
	/* netlink header */
	struct nlmsghdr *nlh = nlmsg_hdr(msg);
	/* generic netlink header*/
	struct genlmsghdr *gnlh = nlmsg_data(nlh);

	if(gnlh->cmd == COLLECTOR_CMD_EVENT) {
		/* we get the attributes*/
		genlmsg_parse(nlh, 0, attrs, COLLECTOR_ATTR_MAX, NULL);
		if (attrs[COLLECTOR_ATTR_ADDR_TRANSMITTER]) {

			struct timeval* now = (struct timeval*)
						nla_data(attrs[COLLECTOR_ATTR_TIMESTAMP]);

			unsigned int event_type = nla_get_u32(attrs[COLLECTOR_ATTR_EVENT]);

			struct mac_address* src = (struct mac_address*) 
					nla_data(attrs[COLLECTOR_ATTR_ADDR_TRANSMITTER]);

			struct mac_address *dst = (struct mac_address*)
					nla_data(attrs[COLLECTOR_ATTR_ADDR_RECEIVER]);

			size_t data_len = nla_get_u32(attrs[COLLECTOR_ATTR_LEN]);

			unsigned int flags = nla_get_u32(attrs[COLLECTOR_ATTR_FLAGS]);

			printf("%ld.%06ld\t", now->tv_sec, now->tv_usec);

			switch(event_type) {
				case EVENT_START_TX: // Transmission started 
				{
					printf("TX STARTED \t");

					printf(" src %02x:%02x:%02x:%02x:%02x:%02x \t", src->addr[0], src->addr[1], src->addr[2], src->addr[3], src->addr[4], src->addr[5]);
					printf(" dst %02x:%02x:%02x:%02x:%02x:%02x \t", dst->addr[0], dst->addr[1], dst->addr[2], dst->addr[3], dst->addr[4], dst->addr[5]);

					printf(" len %u ", data_len);

					printf("\n");

					break;
				}
				case EVENT_END_TX: // Transmission ended
				{
				        int retry_count = -1;
				        int rates_idx, i;
					int success;

					success = !!(flags & IEEE80211_TX_STAT_ACK);

					struct ieee80211_tx_rate* rates = nla_data( attrs[COLLECTOR_ATTR_TX_INFO]);

					for (i = 0; i < IEEE80211_MAX_RATES_PER_TX; i++) {
						if (rates[i].idx < 0) {
							break;
					     	}
				                retry_count += rates[i].count;
					}
					rates_idx = i - 1;

					if( retry_count < 0 )
						retry_count = 0;


					printf("TX ENDED ");				

					if(success)
						printf(" S \t");
					else 
						printf(" F \t");

					printf(" src %02x:%02x:%02x:%02x:%02x:%02x \t", src->addr[0], src->addr[1], src->addr[2], src->addr[3], src->addr[4], src->addr[5]);
					printf(" dst %02x:%02x:%02x:%02x:%02x:%02x \t", dst->addr[0], dst->addr[1], dst->addr[2], dst->addr[3], dst->addr[4], dst->addr[5]);

					printf(" len %u \t", data_len);

					printf(" retry %d \t rate %d \n", retry_count,rates[rates_idx].idx);

					break;
				}
				case EVENT_END_RX: // Reception ended
				{
					
					int signal = nla_get_u32(attrs[COLLECTOR_ATTR_SIGNAL]);

					int rate = nla_get_u32(attrs[COLLECTOR_ATTR_RX_RATE]);
					
					printf("RX ENDED \t");				

					printf(" src %02x:%02x:%02x:%02x:%02x:%02x \t", src->addr[0], src->addr[1], src->addr[2], src->addr[3], src->addr[4], src->addr[5]);
					printf(" dst %02x:%02x:%02x:%02x:%02x:%02x \t", dst->addr[0], dst->addr[1], dst->addr[2], dst->addr[3], dst->addr[4], dst->addr[5]);

					printf(" len %u \t", data_len);

					printf(" signal %d \t rate %d \n", signal, rate);

					break;
				}
	
			}
			
		}
	}
}

/*
 * 	Send a register message to kernel
 */

int send_register_msg()
{

	msg = nlmsg_alloc();
	if (!msg) {
		printf("Error allocating new message MSG!\n");
		return -1;
	}

	genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, genl_family_get_id(family),
		    0, NLM_F_REQUEST, COLLECTOR_CMD_REGISTER, VERSION_NR);
	nl_send_auto_complete(sock,msg);
	nlmsg_free(msg);

	printf("Kernel registration completed\n");

	return 0;

}

/*
 * 	Init netlink
 */

void init_netlink()
{

	cb = nl_cb_alloc(NL_CB_CUSTOM);
	if (!cb) {
		printf("Error allocating netlink callbacks\n");
		exit(EXIT_FAILURE);
	}

	sock = nl_socket_alloc_cb(cb);
	if (!sock) {
		printf("Error allocationg netlink socket\n");
		exit(EXIT_FAILURE);
	}
	
	genl_connect(sock);
	genl_ctrl_alloc_cache(sock, &cache);
	
	family = genl_ctrl_search_by_name(cache, "MAC80211_COLL");

	if (!family) {
		printf("Family MAC80211_COLL not registered\n");
		exit(EXIT_FAILURE);
	}

	nl_cb_set(cb, NL_CB_MSG_IN, NL_CB_CUSTOM, process_messages_cb, NULL);

	printf("Netlink initialization completed\n");

}

/*
 *	Print the CLI help
 */

void print_help(int exval)
{
	printf("wmediumd v%s - a wireless medium simulator\n", VERSION_STR);
	printf("wmediumd [-h] [-V] [-c FILE] [-o FILE]\n\n");

	printf("  -h              print this help and exit\n");
	printf("  -V              print version and exit\n\n");

	printf("  -c FILE         set intput config file\n");
	printf("  -o FILE         set output config file\n\n");

	exit(exval);
}


int main(int argc, char* argv[]) {

	int opt, ifaces;

	/* no arguments given */
	/*if(argc == 1) {
		fprintf(stderr, "This program needs arguments....\n\n");
		print_help(EXIT_FAILURE);
	}*/

	while((opt = getopt(argc, argv, "hV")) != -1) {
		switch(opt) {
		case 'h':
			print_help(EXIT_SUCCESS);
    			break;
		case 'V':
			printf("wcollector v%s - event collector "
			       "for mac80211\n", VERSION_STR);
			exit(EXIT_SUCCESS);
			break;
		case ':':
			printf("wcollector: Error - Option `%c' "
			       "needs a value\n\n", optopt);
			print_help(EXIT_FAILURE);
			break;
		case '?':
			printf("wcollector: Error - No such option:"
			       " `%c'\n\n", optopt);
			print_help(EXIT_FAILURE);
			break;
		}

	}

	if (optind < argc)
		print_help(EXIT_FAILURE);

	/*init netlink*/
	init_netlink();

	/*register userspace daemon*/
	send_register_msg();

	/* Wait for incoming msg*/
	while(1) {
		nl_recvmsgs_default(sock);
	}
	return EXIT_SUCCESS;
}
